#%%

def dijkstra_queue(graph: dict, node: str) -> dict:
    """
    This implementation uses a queue to store the nodes to be visited.
    The algorithm returns the shortest distance from the given node to all other nodes.
    """
    distances = {node: float('infinity') for node in graph}
    distances[node] = 0
    queue = [node]

    while queue:
        current = queue.pop(0)

        for neighbor, weight in graph[current].items():
            new_route = distances[current] + weight

            if new_route < distances[neighbor]:
                distances[neighbor] = new_route
                queue.append(neighbor)

    return distances

graph = {
  "n1": { "n2": 1, "n3": 5 },
  "n2": { "n1": 1, "n3": 7 },
  "n3": { "n1": 5, "n2": 7, "n4": 8 },
  "n4": { "n3": 8 }
}

distances = dijkstra_queue(graph, "n2")
print(distances)


#%% A more efficient implementation using a priority queue

import heapq

def dijkstra_prioqueue(graph: dict, node: str) -> dict:
    """
    This implemententation uses a priority queue to store the nodes to be visited.
    The algorithm returns the shortest distance from the given node to all other nodes.
    """
    distances = {node: float('infinity') for node in graph}
    distances[node] = 0
    pq = [(0, node)] # Tuple (distance, node)

    while len(pq) > 0:
        current_distance, current_node = heapq.heappop(pq)
        if current_distance > distances[current_node]:
            continue

        for neighbor, weight in graph[current_node].items():
            distance = current_distance + weight
            
            if distance < distances[neighbor]:
                distances[neighbor] = distance
                heapq.heappush(pq, (distance, neighbor))
    return distances

graph = {
  "n1": { "n2": 1, "n3": 5 },
  "n2": { "n1": 1, "n3": 7 },
  "n3": { "n1": 5, "n2": 7, "n4": 8 },
  "n4": { "n3": 8 }
}

print(dijkstra_prioqueue(graph, "n2"))

#%%

import heapq

def dijkstra_return_path_priority_queue(graph, start, goal):
    """
    This implemententation uses a priority queue to store the nodes to be visited.
    The algorithm also builds a list of the shortest path from the start node to the goal node.
    """
    distances = {node: float('infinity') for node in graph}
    distances[start] = 0
    pq = [(0, start)] # Tuple (distance, node)

    previous_nodes = {node: None for node in graph}

    while len(pq) > 0:
        current_distance, current_node = heapq.heappop(pq)
        if current_distance > distances[current_node]:
            continue

        for neighbor, weight in graph[current_node].items():
            distance = current_distance + weight
            
            if distance < distances[neighbor]:
                distances[neighbor] = distance
                previous_nodes[neighbor] = current_node
                heapq.heappush(pq, (distance, neighbor))

    # start at goal and work back
    path, current_node = [], goal
    while previous_nodes[current_node] is not None:
        path.insert(0, current_node)
        print("current_node: ", current_node)
        current_node = previous_nodes[current_node]
    
    if path:
        print("current_node: ", current_node)
        path.insert(0, current_node)
    return path

graph = {
  "n1": { "n2": 1, "n3": 5 },
  "n2": { "n1": 1, "n3": 7 },
  "n3": { "n1": 5, "n2": 7, "n4": 8 },
  "n4": { "n3": 8 }
}

print(dijkstra_return_path_priority_queue(graph, "n2", "n4"))
# %%
